import java.text.SimpleDateFormat;
import java.time.Duration;
import java.util.Date;
import java.util.HashMap;
import java.util.concurrent.Callable;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

import javax.jms.Connection;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageProducer;
import javax.jms.ObjectMessage;
import javax.jms.Session;
import javax.jms.TextMessage;

import org.hornetq.api.core.TransportConfiguration;
import org.hornetq.api.jms.HornetQJMSClient;
import org.hornetq.api.jms.JMSFactoryType;
import org.hornetq.core.remoting.impl.netty.NettyConnectorFactory;
import org.hornetq.core.remoting.impl.netty.TransportConstants;
import org.hornetq.jms.client.HornetQConnectionFactory;

public class QueueExampleForDiscovery {

    public static void main(String[] args) throws InterruptedException, JMSException {
        new QueueExampleForDiscovery().run();
    }

    private static final SimpleDateFormat DATE_FORMAT = new SimpleDateFormat(
            "dd.MM.yyyy HH:mm:ss;SSS");

    private final ExecutorService executorService = Executors.newCachedThreadPool();

    private Connection connection;

    private final int numberOfMessages = Integer.MAX_VALUE;

    private final String queueName = "foobar";

    private final CountDownLatch latch = new CountDownLatch(2);

    private HornetQConnectionFactory createStaticFactory() {

        TransportConfiguration[] transportConfiguration = createTransportConfiguration();

        HornetQConnectionFactory factory = HornetQJMSClient.createConnectionFactoryWithHA(
                JMSFactoryType.CF, transportConfiguration);

        setupFactory(factory);

        return factory;

    }

    private void setupFactory(HornetQConnectionFactory factory) {
        // http://docs.jboss.org/hornetq/2.3.0.Final/docs/user-manual/html/ha.html#ha.automatic.failover
        factory.setClientFailureCheckPeriod(Duration.ofSeconds(1).toMillis());

        // 39.2.1.1 on
        // http://docs.jboss.org/hornetq/2.3.0.Final/docs/user-manual/html/ha.html#ha.automatic.failover
        factory.setInitialConnectAttempts(5);

        // 34.1
        factory.setConfirmationWindowSize(1000000);

        // 34.1
        factory.setReconnectAttempts(20);
        factory.setRetryIntervalMultiplier(1.5);
        factory.setMaxRetryInterval(8000);
    }

    private TransportConfiguration[] createTransportConfiguration() {
        return new TransportConfiguration[] { createTransportConfiguration("localhost", 62500),
                createTransportConfiguration("localhost", 62510),
        // createTransportConfiguration("localhost", 62520)
        };
    }

    private TransportConfiguration createTransportConfiguration(String host, int port) {
        HashMap<String, Object> map = new HashMap<String, Object>();
        map.put(TransportConstants.HOST_PROP_NAME, host);
        map.put(TransportConstants.PORT_PROP_NAME, String.valueOf(port));
        TransportConfiguration tc = new TransportConfiguration(NettyConnectorFactory.class
                .getName(), map);
        return tc;
    }

    public void run() throws InterruptedException, JMSException {

        try {
            connection = createStaticFactory().createConnection();
            connection.start();

            startConsumer();

            startProducer();

            latch.await();

            executorService.shutdown();
            executorService.awaitTermination(15, TimeUnit.SECONDS);
        } finally {
            if (connection != null) {
                connection.stop();
                connection.close();
            }
            executorService.shutdownNow();
        }
    }

    private void startProducer() {

        Callable<Void> task = new Callable<Void>() {

            @Override
            public Void call() {
                try {
                    produce();
                } catch (Exception e) {
                    e.printStackTrace();
                }
                return null;
            }

            private void produce() throws JMSException {
                Session session = connection.createSession(true, Session.AUTO_ACKNOWLEDGE);
                Destination destination = session.createQueue(queueName);
                MessageProducer producer = session.createProducer(destination);

                for (int i = 0; i < numberOfMessages; i++) {
                    String text = i + " [" + DATE_FORMAT.format(new Date()) + "]";
                    TextMessage message = session.createTextMessage(text);
                    System.out.println("sending message " + text);
                    producer.send(message);
                    session.commit();
                }

                producer.close();
                session.close();

                latch.countDown();

            }

        };

        executorService.submit(task);

    }

    private void startConsumer() {

        Callable<Void> task = new Callable<Void>() {

            @Override
            public Void call() {
                try {
                    consume();
                } catch (Exception e) {
                    e.printStackTrace();
                }
                return null;
            }

            private void consume() throws JMSException {
                Session session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
                Destination destination = session.createQueue(queueName);
                javax.jms.MessageConsumer consumer = session.createConsumer(destination);

                while (true) {
                    Message message = consumer.receive(1000L);
                    if (message == null) {
                        break;
                    }
                    if (message instanceof TextMessage) {
                        System.out.println("recieved: " + ((TextMessage) message).getText());
                    } else if (message instanceof ObjectMessage) {
                        System.out.println("recieved: " + ((ObjectMessage) message).getObject());
                    } else {
                        System.out.println("recieved: " + message);
                    }
                }

                consumer.close();
                session.close();

                latch.countDown();

            }

        };

        executorService.submit(task);

    }

}
